<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"
      lang="en">

  <head>
    <title>Test</title>
  </head>

  <body>
    <h1>Test</h1>
    <p>$message</p>
  </body>

</html>
